import asyncio
import functools
import inspect
import time
from typing import Any


def measure(*, reset: bool = False, ms: bool = False):
    global t
    if 't' in globals() and not reset:
        dt = time.time() - t
        line = inspect.stack()[1][0].f_lineno
        output = f'{dt * 1000:7.3f} ms' if ms else f'{dt:7.3f} s'
        print(f'{inspect.stack()[1].filename}:{line}', output, flush=True)
    if reset:
        print('------------', flush=True)
    t = time.time()


def is_coroutine(object: Any) -> bool:
    while isinstance(object, functools.partial):
        object = object.func
    return asyncio.iscoroutinefunction(object)



from typing import Callable, Dict, Optional

import justpy as jp

from .float_element import FloatElement


class Number(FloatElement):

    def __init__(
            self, label: str = None, *,
            placeholder: str = None, value: float = None, format: str = None, on_change: Optional[Callable] = None):
        """Number Input

        :param label: displayed name for the number input
        :param placeholder: text to show if no value is entered
        :param value: the initial value of the field
        :param format: a string like '%.2f' to format the displayed value
        :param on_change: callback to execute when the input is confirmed by leaving the focus
        """
        view = jp.QInput(
            type='number',
            label=label,
            placeholder=placeholder,
            change=self.handle_change,
            disable_input_event=True,
            temp=False,
        )

        super().__init__(view, value=value, format=format, on_change=on_change)

    def handle_change(self, msg: Dict):
        msg['value'] = float(msg['value'])

        return super().handle_change(msg)



from typing import Callable, Optional

import justpy as jp

from .string_element import StringElement


class ColorInput(StringElement):

    def __init__(self, label: str = None, *,
                 placeholder: str = None, value: str = '', on_change: Optional[Callable] = None):
        """Color Input

        :param label: displayed label for the color input
        :param placeholder: text to show if no color is selected
        :param value: the current color value
        :param on_change: callback to execute when the input is confirmed by leaving the focus
        """
        view = jp.QInput(
            label=label,
            placeholder=placeholder,
            value=value,
            change=self.handle_change,
            disable_input_event=True,
            temp=False,
        )

        self._icon_button = jp.parse_html('''
            <q-icon name="colorize" class="cursor-pointer">
                <q-popup-proxy transition-show="scale" transition-hide="scale" name="popup" no-parent-event>
                    <q-color name="color_input"/>
                </q-popup-proxy>
            </q-icon>''')
        view.add_scoped_slot('append', self._icon_button)
        self._icon_button.on('click', lambda *_: self.open() or False)
        self._icon_button.name_dict['color_input'].on('change', self.handle_change)
        self._icon_button.name_dict['color_input'].disable_input_event = True
        self._icon_button.name_dict['popup'].on('input', lambda *_: self.update() or False)
        self._icon_button.name_dict['popup'].on('show', lambda *_: self.update() or False)
        self._icon_button.name_dict['popup'].on('hide', lambda *_: self.update() or False)

        super().__init__(view, value=value, on_change=on_change)

    def open(self):
        self._icon_button.name_dict['popup'].value = True
        self.update()

    def close(self):
        self.view.name_dict['popup'].value = False
        self.update()



import os
from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class Config():
    # NOTE: should be in sync with ui.run arguments
    host: str = os.environ.get('HOST', '0.0.0.0')
    port: int = int(os.environ.get('PORT', '8080'))
    title: str = 'NiceGUI'
    reload: bool = True
    favicon: str = 'favicon.ico'
    dark: Optional[bool] = False
    main_page_classes: str = 'q-ma-md column items-start'
    binding_refresh_interval: float = 0.1
    excludes: List[str] = field(default_factory=list)



from typing import Callable, Optional

import justpy as jp

from ..events import ClickEventArguments, handle_event
from ..task_logger import create_task
from .element import Element


class MenuItem(Element):

    def __init__(self, text: str = '', on_click: Optional[Callable] = None, *, auto_close: bool = True):
        """Menu Item

        A menu item to be added to a menu.

        :param text: label of the menu item
        :param on_click: callback to be executed when selecting the menu item
        :param auto_close: whether the menu should be closed after a click event (default: `True`)
        """
        view = jp.QItem(text=text, clickable=True, temp=False)

        def handle_click(view, event) -> Optional[bool]:
            result = handle_event(on_click, ClickEventArguments(sender=self, socket=event.get('websocket')))
            if auto_close:
                assert isinstance(self.parent_view, jp.QMenu)
                self.parent_view.value = False
                create_task(self.parent_view.update())
            return result

        view.on('click', handle_click)

        super().__init__(view)



from __future__ import annotations

import traceback
from typing import Any, Callable, Dict, List, Optional

from justpy import WebPage

from ..events import MouseEventArguments, handle_event
from ..routes import add_dependencies
from .custom_view import CustomView
from .element import Element

add_dependencies(__file__)


class InteractiveImageView(CustomView):

    def __init__(self, source: str, on_mouse: Callable, events: List[str], cross: bool):
        super().__init__('interactive_image', source=source, events=events, cross=cross, svg_content='')
        self.allowed_events = ['onMouse', 'onConnect']
        self.initialize(onMouse=on_mouse, onConnect=self.on_connect)
        self.sockets = []

    def on_connect(self, msg):
        self.prune_sockets()
        self.sockets.append(msg.websocket)

    def prune_sockets(self):
        page_sockets = [s for page_id in self.pages for s in WebPage.sockets.get(page_id, {}).values()]
        self.sockets = [s for s in self.sockets if s in page_sockets]


class InteractiveImage(Element):

    def __init__(self, source: str, *,
                 on_mouse: Optional[Callable] = None, events: List[str] = ['click'], cross: bool = False):
        """Interactive Image

        Create an image with an SVG overlay that handles mouse events and yields image coordinates.

        :param source: the source of the image; can be an URL or a base64 string
        :param on_mouse: callback for mouse events (yields `type`, `image_x` and `image_y`)
        :param events: list of JavaScript events to subscribe to (default: `['click']`)
        :param cross: whether to show crosshairs (default: `False`)
        """
        self.mouse_handler = on_mouse
        super().__init__(InteractiveImageView(source, self.handle_mouse, events, cross))

    def handle_mouse(self, msg: Dict[str, Any]) -> Optional[bool]:
        if self.mouse_handler is None:
            return False
        try:
            arguments = MouseEventArguments(
                sender=self,
                socket=msg.get('websocket'),
                type=msg.get('mouse_event_type'),
                image_x=msg.get('image_x'),
                image_y=msg.get('image_y'),
            )
            return handle_event(self.mouse_handler, arguments)
        except:
            traceback.print_exc()

    async def set_source(self, source: str):
        self.view.options.source = source
        self.view.prune_sockets()
        for socket in self.view.sockets:
            await self.view.run_method(f'set_source("{source}")', socket)

    @property
    def svg_content(self) -> str:
        return self.view.options.svg_content

    @svg_content.setter
    def svg_content(self, content: str):
        self.view.options.svg_content = content



from typing import Dict

import justpy as jp

from .element import Element


class Table(Element):

    def __init__(self, options: Dict):
        """Table

        An element to create a table using `AG Grid <https://www.ag-grid.com/>`_.

        :param options: dictionary of AG Grid options
        """
        view = jp.AgGrid(temp=False)
        view.options = self.options = jp.Dict(**options)
        super().__init__(view)



from __future__ import annotations

import asyncio
import logging
from contextlib import contextmanager
from typing import TYPE_CHECKING, Awaitable, Callable, Dict, Generator, List, Optional, Union

from starlette.applications import Starlette
from uvicorn import Server

from .config import Config
from .task_logger import create_task

if TYPE_CHECKING:
    import justpy as jp

    from .page_builder import PageBuilder

app: Starlette
config: Optional[Config] = None
server: Optional[Server] = None
loop: Optional[asyncio.AbstractEventLoop] = None
page_builders: Dict[str, 'PageBuilder'] = {}
view_stacks: Dict[List['jp.HTMLBaseComponent']] = {}
tasks: List[asyncio.tasks.Task] = []
log: logging.Logger = logging.getLogger('nicegui')
connect_handlers: List[Union[Callable, Awaitable]] = []
disconnect_handlers: List[Union[Callable, Awaitable]] = []
startup_handlers: List[Union[Callable, Awaitable]] = []
shutdown_handlers: List[Union[Callable, Awaitable]] = []


def find_route(function: Callable) -> str:
    routes = [route for route, page_builder in page_builders.items() if page_builder.function == function]
    if not routes:
        raise ValueError(f'Invalid page function {function}')
    return routes[0]


def get_task_id() -> int:
    return id(asyncio.current_task()) if loop and loop.is_running() else 0


def get_view_stack() -> List['jp.HTMLBaseComponent']:
    task_id = get_task_id()
    if task_id not in view_stacks:
        view_stacks[task_id] = []
    return view_stacks[task_id]


def prune_view_stack() -> None:
    task_id = get_task_id()
    if not view_stacks[task_id]:
        del view_stacks[task_id]


@contextmanager
def within_view(view: 'jp.HTMLBaseComponent') -> Generator[None, None, None]:
    child_count = len(view)
    get_view_stack().append(view)
    yield
    get_view_stack().pop()
    prune_view_stack()
    if len(view) != child_count:
        create_task(view.update())



import asyncio
import logging
import time
from collections import defaultdict
from typing import Any, Callable, Optional, Set, Tuple

from justpy.htmlcomponents import HTMLBaseComponent

from . import globals
from .task_logger import create_task

bindings = defaultdict(list)
bindable_properties = dict()
active_links = []


async def loop():
    while True:
        visited: Set[Tuple[int, str]] = set()
        visited_views: Set[HTMLBaseComponent] = set()
        t = time.time()
        for link in active_links:
            (source_obj, source_name, target_obj, target_name, transform) = link
            value = transform(getattr(source_obj, source_name))
            if getattr(target_obj, target_name) != value:
                setattr(target_obj, target_name, value)
                propagate(target_obj, target_name, visited, visited_views)
        if time.time() - t > 0.01:
            logging.warning(f'binding propagation for {len(active_links)} active links took {time.time() - t:.3f} s')
        t = time.time()
        update_views(visited_views)
        if time.time() - t > 0.01:
            logging.warning(f'binding update for {len(visited_views)} visited views took {time.time() - t:.3f} s')
        await asyncio.sleep(globals.config.binding_refresh_interval)


async def update_views_async(views: Set[HTMLBaseComponent]):
    for view in views:
        await view.update()


def update_views(views: Set[HTMLBaseComponent]):
    if globals.loop is None:
        return  # NOTE: no need to update view if event loop is not running, yet
    create_task(update_views_async(views), name='update_views_async')


def propagate(source_obj,
              source_name,
              visited: Set[Tuple[int, str]] = None,
              visited_views: Set[HTMLBaseComponent] = None) -> Set[HTMLBaseComponent]:
    if visited is None:
        visited = set()
    if visited_views is None:
        visited_views = set()
    visited.add((id(source_obj), source_name))
    if isinstance(source_obj, HTMLBaseComponent):
        visited_views.add(source_obj)
    for _, target_obj, target_name, transform in bindings[(id(source_obj), source_name)]:
        if (id(target_obj), target_name) in visited:
            continue
        target_value = transform(getattr(source_obj, source_name))
        if getattr(target_obj, target_name) != target_value:
            setattr(target_obj, target_name, target_value)
            propagate(target_obj, target_name, visited, visited_views)
    return visited_views


def bind_to(self_obj: Any, self_name: str, other_obj: Any, other_name: str, forward: Callable):
    bindings[(id(self_obj), self_name)].append((self_obj, other_obj, other_name, forward))
    if (id(self_obj), self_name) not in bindable_properties:
        active_links.append((self_obj, self_name, other_obj, other_name, forward))
    update_views(propagate(self_obj, self_name))


def bind_from(self_obj: Any, self_name: str, other_obj: Any, other_name: str, backward: Callable):
    bindings[(id(other_obj), other_name)].append((other_obj, self_obj, self_name, backward))
    if (id(other_obj), other_name) not in bindable_properties:
        active_links.append((other_obj, other_name, self_obj, self_name, backward))
    update_views(propagate(other_obj, other_name))


class BindableProperty:

    def __init__(self, on_change: Optional[Callable] = None):
        self.on_change = on_change

    def __set_name__(self, _, name: str):
        self.name = name

    def __get__(self, owner: Any, _=None):
        return getattr(owner, '_' + self.name)

    def __set__(self, owner: Any, value: Any):
        has_attribute = hasattr(owner, '_' + self.name)
        value_changed = has_attribute and getattr(owner, '_' + self.name) != value
        if has_attribute and not value_changed:
            return
        setattr(owner, '_' + self.name, value)
        bindable_properties[(id(owner), self.name)] = owner
        update_views(propagate(owner, self.name))
        if value_changed and self.on_change is not None:
            self.on_change(owner, value)



import traceback
from typing import Callable, Dict, Optional

from ..events import KeyboardAction, KeyboardKey, KeyboardModifiers, KeyEventArguments, handle_event
from ..routes import add_dependencies
from .custom_view import CustomView
from .element import Element

add_dependencies(__file__)


class KeyboardView(CustomView):

    def __init__(self, on_key: Callable, repeating: bool):
        super().__init__('keyboard', active_js_events=['keydown', 'keyup'], repeating=repeating)
        self.allowed_events = ['keyboardEvent']
        self.style = 'display: none'
        self.initialize(temp=False, on_keyboardEvent=on_key)


class Keyboard(Element):

    def __init__(self, *, on_key: Optional[Callable] = None, active: bool = True, repeating: bool = True):
        """
        Keyboard

        Adds global keyboard event tracking.

        :param on_key: callback to be executed when keyboard events occur.
        :param active: boolean flag indicating whether the callback should be executed or not (default: `True`)
        :param repeating: boolean flag indicating whether held keys should be sent repeatedly (default: `True`)
        """
        super().__init__(KeyboardView(on_key=self.handle_key, repeating=repeating))
        self.active = active
        self.key_handler = on_key

    def handle_key(self, msg: Dict) -> Optional[bool]:
        if not self.active:
            return False

        try:
            action = KeyboardAction(
                keydown=msg.key_data.action == 'keydown',
                keyup=msg.key_data.action == 'keyup',
                repeat=msg.key_data.repeat,
            )
            modifiers = KeyboardModifiers(
                alt=msg.key_data.altKey,
                ctrl=msg.key_data.ctrlKey,
                meta=msg.key_data.metaKey,
                shift=msg.key_data.shiftKey,
            )
            key = KeyboardKey(
                name=msg.key_data.key,
                code=msg.key_data.code,
                location=msg.key_data.location,
            )
            arguments = KeyEventArguments(
                sender=self,
                socket=msg.websocket,
                action=action,
                modifiers=modifiers,
                key=key
            )
            return handle_event(self.key_handler, arguments)
        except Exception:
            traceback.print_exc()



import justpy as jp

from .group import Group


class Menu(Group):

    def __init__(self, *, value: bool = False):
        """Menu

        Creates a menu.

        :param value: whether the menu is already opened (default: `False`)
        """
        view = jp.QMenu(value=value, temp=False, no_parent_event=True)

        super().__init__(view)

    def open(self):
        self.view.value = True
        self.update()

    def close(self):
        self.view.value = False
        self.update()



from typing import Awaitable, Callable, Union

import justpy as jp

from . import globals


def on_connect(self, handler: Union[Callable, Awaitable]):
    globals.connect_handlers.append(handler)


def on_disconnect(self, handler: Union[Callable, Awaitable]):
    globals.disconnect_handlers.append(handler)


def on_startup(self, handler: Union[Callable, Awaitable]):
    globals.startup_handlers.append(handler)


def on_shutdown(self, handler: Union[Callable, Awaitable]):
    globals.shutdown_handlers.append(handler)


async def shutdown(self) -> None:
    if globals.config.reload:
        raise Exception('ui.shutdown is not supported when auto-reload is enabled')
    for socket in [s for page in jp.WebPage.sockets.values() for s in page.values()]:
        await socket.close()
    globals.server.should_exit = True



from typing import Any, Callable, Dict, List, Optional, Union

import justpy as jp

from .choice_element import ChoiceElement


class Select(ChoiceElement):

    def __init__(self, options: Union[List, Dict], *,
                 label: Optional[str] = None, value: Any = None, on_change: Optional[Callable] = None):
        """Dropdown Selection

        :param options: a list ['value1', ...] or dictionary `{'value1':'label1', ...}` specifying the options
        :param value: the initial value
        :param on_change: callback to execute when selection changes
        """
        view = jp.QSelect(options=options, label=label, input=self.handle_change, temp=False)

        super().__init__(view, options, value=value, on_change=on_change)

    def value_to_view(self, value: Any):
        matches = [o for o in self.view.options if o['value'] == value]
        if any(matches):
            return matches[0]['label']
        else:
            return value

    def handle_change(self, msg: Dict):
        msg['label'] = msg['value']['label']
        msg['value'] = msg['value']['value']
        return super().handle_change(msg)



import justpy as jp

from ..task_logger import create_task
from .element import Element


class Notify(Element):

    def __init__(self, message: str, *, position: str = 'bottom', close_button: str = None):
        """Notification

        Displays a notification on the screen.

        :param message: content of the notification
        :param position: position on the screen ("top-left", "top-right", "bottom-left","bottom-right, "top", "bottom", "left", "right" or "center", default: "bottom")
        :param close_button: optional label of a button to dismiss the notification (default: `None`)
        """
        view = jp.QNotify(message=message, position=position, closeBtn=close_button, temp=False)

        super().__init__(view)
        self.visible = False
        create_task(self.notify_async(), name='notify_async')

    async def notify_async(self):
        self.view.notify = True
        await self.parent_view.update()
        self.view.notify = False
        self.parent_view.remove_component(self.view)



from ..routes import add_dependencies
from .custom_view import CustomView
from .element import Element

add_dependencies(__file__)


class ColorsView(CustomView):

    def __init__(self, primary, secondary, accent, positive, negative, info, warning):
        super().__init__('colors',
                         primary=primary,
                         secondary=secondary,
                         accent=accent,
                         positive=positive,
                         negative=negative,
                         info=info,
                         warning=warning)
        self.initialize(temp=False)


class Colors(Element):

    def __init__(self, *,
                 primary='#1976d2',
                 secondary='#26a69a',
                 accent='#9c27b0',
                 positive='#21ba45',
                 negative='#c10015',
                 info='#31ccec',
                 warning='#f2c037'):
        """Color Theming

        Sets the main colors (primary, secondary, accent, ...) used by `Quasar <https://quasar.dev/>`_.
        """
        super().__init__(ColorsView(primary, secondary, accent, positive, negative, info, warning))

        self.update()



from __future__ import annotations

from typing import List, Optional

import numpy as np
from justpy import WebPage

from .scene_object3d import Object3D


class Scene(Object3D):

    def __init__(self, view):
        super().__init__('scene', view)


class Group(Object3D):

    def __init__(self):
        super().__init__('group')


class Box(Object3D):

    def __init__(self,
                 width: float = 1.0,
                 height: float = 1.0,
                 depth: float = 1.0,
                 wireframe: bool = False,
                 ):
        super().__init__('box', width, height, depth, wireframe)


class Sphere(Object3D):

    def __init__(self,
                 radius: float = 1.0,
                 width_segments: int = 32,
                 height_segments: int = 16,
                 wireframe: bool = False,
                 ):
        super().__init__('sphere', radius, width_segments, height_segments, wireframe)


class Cylinder(Object3D):

    def __init__(self,
                 top_radius: float = 1.0,
                 bottom_radius: float = 1.0,
                 height: float = 1.0,
                 radial_segments: int = 8,
                 height_segments: int = 1,
                 wireframe: bool = False,
                 ):
        super().__init__('cylinder', top_radius, bottom_radius, height, radial_segments, height_segments, wireframe)


class Ring(Object3D):

    def __init__(self,
                 inner_radius: float = 0.5,
                 outer_radius: float = 1.0,
                 theta_segments: int = 8,
                 phi_segments: int = 1,
                 theta_start: float = 0,
                 theta_length: float = 2 * np.pi,
                 wireframe: bool = False,
                 ):
        super().__init__('ring',
                         inner_radius, outer_radius, theta_segments, phi_segments, theta_start, theta_length, wireframe)


class QuadraticBezierTube(Object3D):

    def __init__(self,
                 start: List(float, float, float),
                 mid: List(float, float, float),
                 end: List(float, float, float),
                 tubular_segments: int = 64,
                 radius: float = 1.0,
                 radial_segments: int = 8,
                 closed: bool = False,
                 wireframe: bool = False,
                 ):
        super().__init__('quadratic_bezier_tube',
                         start, mid, end, tubular_segments, radius, radial_segments, closed, wireframe)


class Extrusion(Object3D):

    def __init__(self,
                 outline: List[List[float, float]],
                 height: float,
                 wireframe: bool = False,
                 ):
        super().__init__('extrusion', outline, height, wireframe)


class Stl(Object3D):

    def __init__(self,
                 url: str,
                 wireframe: bool = False,
                 ):
        super().__init__('stl', url, wireframe)


class Line(Object3D):

    def __init__(self,
                 start: List[float, float, float],
                 end: List[float, float, float],
                 ):
        super().__init__('line', start, end)


class Curve(Object3D):

    def __init__(self,
                 start: List[float, float, float],
                 control1: List[float, float, float],
                 control2: List[float, float, float],
                 end: List[float, float, float],
                 num_points: int = 20,
                 ):
        super().__init__('curve', start, control1, control2, end, num_points)


class Text(Object3D):

    def __init__(self,
                 text: str,
                 style: str = '',
                 ):
        super().__init__('text', text, style)


class Text3d(Object3D):

    def __init__(self,
                 text: str,
                 style: str = '',
                 ):
        super().__init__('text3d', text, style)


class Texture(Object3D):

    def __init__(self,
                 url: str,
                 coordinates: List[List[Optional[List[float]]]],
                 ):
        super().__init__('texture', url, coordinates)

    async def set_url(self, url: str):
        self.args[0] = url
        for socket in WebPage.sockets.get(self.page.page_id, {}).values():
            await self.view.run_method(f'set_texture_url("{self.id}", "{url}")', socket)

    async def set_coordinates(self, coordinates: List[List[Optional[List[float]]]]):
        self.args[1] = coordinates
        for socket in WebPage.sockets.get(self.page.page_id, {}).values():
            await self.view.run_method(f'set_texture_coordinates("{self.id}", {coordinates})', socket)


class SpotLight(Object3D):

    def __init__(self,
                 color: str = '#ffffff',
                 intensity: float = 1.0,
                 distance: float = 0.0,
                 angle: float = np.pi / 3,
                 penumbra: float = 0.0,
                 decay: float = 1.0,
                 ):
        super().__init__('spot_light', color, intensity, distance, angle, penumbra, decay)



import inspect
import os.path
from functools import wraps
from typing import List

import justpy as jp
from starlette.requests import Request
from starlette.responses import FileResponse, PlainTextResponse
from starlette.routing import BaseRoute, Mount, Route, compile_path
from starlette.staticfiles import StaticFiles

from . import globals
from .helpers import is_coroutine


def add_route(self, route: BaseRoute) -> None:
    """
    :param route: starlette route including a path and a function to be called
    :return:
    """
    globals.app.routes.insert(0, route)


def add_static_files(self, path: str, directory: str) -> None:
    """
    :param path: string that starts with a '/'
    :param directory: folder with static files to serve under the given path
    """
    add_route(None, Mount(path, app=StaticFiles(directory=directory)))


def get(self, path: str):
    """
    Use as a decorator for a function like @ui.get('/another/route/{id}').
    :param path: string that starts with a '/'
    :return:
    """
    *_, converters = compile_path(path)

    def decorator(func):
        @wraps(func)
        async def decorated(request: Request):
            args = {name: converter.convert(request.path_params.get(name)) for name, converter in converters.items()}
            parameters = inspect.signature(func).parameters
            for key in parameters:
                if parameters[key].annotation.__name__ == 'bool':
                    args[key] = bool(args[key])
                if parameters[key].annotation.__name__ == 'int':
                    args[key] = int(args[key])
                elif parameters[key].annotation.__name__ == 'float':
                    args[key] = float(args[key])
                elif parameters[key].annotation.__name__ == 'complex':
                    args[key] = complex(args[key])
            if 'request' in parameters and 'request' not in args:
                args['request'] = request
            return await func(**args) if is_coroutine(func) else func(**args)
        self.add_route(Route(path, decorated))
        return decorated
    return decorator


def add_dependencies(py_filepath: str, dependencies: List[str] = []) -> None:
    vue_filepath = os.path.splitext(os.path.realpath(py_filepath))[0] + '.js'

    for dependency in dependencies:
        is_remote = dependency.startswith('http://') or dependency.startswith('https://')
        src = dependency if is_remote else f'lib/{dependency}'
        if src not in jp.component_file_list:
            jp.component_file_list += [src]
            if not is_remote:
                filepath = f'{os.path.dirname(vue_filepath)}/{src}'
                route = Route(f'/{src}', lambda _, filepath=filepath: FileResponse(filepath))
                jp.app.routes.insert(0, route)

    if vue_filepath not in jp.component_file_list:
        filename = os.path.basename(vue_filepath)
        jp.app.routes.insert(0, Route(f'/{filename}', lambda _: FileResponse(vue_filepath)))
        jp.component_file_list += [filename]


def create_exclude_routes() -> None:
    def void(_: Request) -> PlainTextResponse:
        return PlainTextResponse()

    if 'chart' in globals.config.excludes:
        add_route(None, Route('/templates/local/highcharts.js', void))
    if 'colors' in globals.config.excludes:
        add_route(None, Route('/colors.js', void))
    if 'custom_example' in globals.config.excludes:
        add_route(None, Route('/custom_example.js', void))
    if 'interactive_image' in globals.config.excludes:
        add_route(None, Route('/interactive_image.js', void))
    if 'keyboard' in globals.config.excludes:
        add_route(None, Route('/keyboard.js', void))
    if 'log' in globals.config.excludes:
        add_route(None, Route('/log.js', void))
    if 'joystick' in globals.config.excludes:
        add_route(None, Route('/joystick.js', void))
        add_route(None, Route('/lib/nipplejs.min.js', void))
    if 'scene' in globals.config.excludes:
        add_route(None, Route('/scene.js', void))
        add_route(None, Route('/lib/CSS2DRenderer.js', void))
        add_route(None, Route('/lib/CSS3DRenderer.js', void))
        add_route(None, Route('/lib/OrbitControls.js', void))
        add_route(None, Route('/lib/STLLoader.js', void))
        add_route(None, Route('/lib/three.min.js', void))
        add_route(None, Route('/lib/tween.umd.min.js', void))
    if 'table' in globals.config.excludes:
        add_route(None, Route('/templates/local/ag-grid-community.js', void))



from typing import Callable, Optional

import justpy as jp

from .bool_element import BoolElement


class Switch(BoolElement):

    def __init__(self, text: str = '', *, value: bool = False, on_change: Optional[Callable] = None):
        """Switch

        :param text: the label to display next to the switch
        :param value: whether it should be active initially (default: `False`)
        :param on_change: callback which is invoked when state is changed by the user
        """
        view = jp.QToggle(text=text, input=self.handle_change, temp=False)

        super().__init__(view, value=value, on_change=on_change)



from typing import Optional

import justpy as jp

from ..binding import BindableProperty, bind_from, bind_to
from .group import Group


class Expansion(Group):
    text = BindableProperty()

    def __init__(self, text: str, *, icon: Optional[str] = None):
        '''Expansion Element

        Provides an expandable container.

        :param text: title text
        :param icon: optional icon (default: None)
        '''
        view = jp.QExpansionItem(label=text, icon=icon, delete_flag=False, temp=False)
        super().__init__(view)

        self.text = text
        self.bind_text_to(self.view, 'label')

    def set_text(self, text: str):
        self.text = text

    def bind_text_to(self, target_object, target_name, forward=lambda x: x):
        bind_to(self, 'text', target_object, target_name, forward=forward)
        return self

    def bind_text_from(self, target_object, target_name, backward=lambda x: x):
        bind_from(self, 'text', target_object, target_name, backward=backward)
        return self

    def bind_text(self, target_object, target_name, forward=lambda x: x, backward=lambda x: x):
        bind_from(self, 'text', target_object, target_name, backward=backward)
        bind_to(self, 'text', target_object, target_name, forward=forward)
        return self



from __future__ import annotations

from typing import Dict, Optional

import justpy as jp

from .. import globals
from ..binding import BindableProperty, bind_from, bind_to
from ..page import Page, find_parent_view
from ..task_logger import create_task


def _handle_visibility_change(sender: Element, visible: bool) -> None:
    (sender.view.remove_class if visible else sender.view.set_class)('hidden')
    sender.update()


class Element:
    visible = BindableProperty(on_change=_handle_visibility_change)

    def __init__(self, view: jp.HTMLBaseComponent):
        self.parent_view = find_parent_view()
        self.parent_view.add(view)
        self.view = view
        assert len(self.parent_view.pages) == 1
        self.page: Page = list(self.parent_view.pages.values())[0]
        self.view.add_page(self.page)

        self.visible = True

    def bind_visibility_to(self, target_object, target_name, forward=lambda x: x):
        bind_to(self, 'visible', target_object, target_name, forward=forward)
        return self

    def bind_visibility_from(self, target_object, target_name, backward=lambda x: x, *, value=None):
        if value is not None:
            def backward(x): return x == value

        bind_from(self, 'visible', target_object, target_name, backward=backward)
        return self

    def bind_visibility(self, target_object, target_name, forward=lambda x: x, backward=None, *, value=None):
        if value is not None:
            def backward(x): return x == value

        bind_from(self, 'visible', target_object, target_name, backward=backward)
        bind_to(self, 'visible', target_object, target_name, forward=forward)
        return self

    def classes(self, add: Optional[str] = None, *, remove: Optional[str] = None, replace: Optional[str] = None):
        '''HTML classes to modify the look of the element.
        Every class in the `remove` parameter will be removed from the element.
        Classes are separated with a blank space.
        This can be helpful if the predefined classes by NiceGUI are not wanted in a particular styling.
        '''
        class_list = self.view.classes.split() if replace is None else []
        class_list = [c for c in class_list if c not in (remove or '').split()]
        class_list += (add or '').split()
        class_list += (replace or '').split()
        new_classes = ' '.join(dict.fromkeys(class_list))  # NOTE: remove duplicates while preserving order
        if self.view.classes != new_classes:
            self.view.classes = new_classes
            self.update()
        return self

    def style(self, add: Optional[str] = None, *, remove: Optional[str] = None, replace: Optional[str] = None):
        '''CSS style sheet definitions to modify the look of the element.
        Every style in the `remove` parameter will be removed from the element.
        Styles are separated with a semicolon.
        This can be helpful if the predefined style sheet definitions by NiceGUI are not wanted in a particular styling.
        '''
        def str_to_dict(s: Optional[str]) -> Dict[str, str]:
            return dict((word.strip() for word in part.split(':')) for part in s.strip('; ').split(';')) if s else {}
        style_dict = str_to_dict((self.view.style or '').strip('; ')) if replace is None else {}
        for key in str_to_dict(remove):
            del style_dict[key]
        style_dict.update(str_to_dict(add))
        style_dict.update(str_to_dict(replace))
        new_style = ';'.join(f'{key}:{value}' for key, value in style_dict.items())
        if self.view.style != new_style:
            self.view.style = new_style
            self.update()
        return self

    def props(self, add: Optional[str] = None, *, remove: Optional[str] = None):
        '''Quasar props https://quasar.dev/vue-components/button#design to modify the look of the element.
        Boolean props will automatically activated if they appear in the list of the `add` property.
        Props are separated with a blank space.
        Every prop passed to the `remove` parameter will be removed from the element.
        This can be helpful if the predefined props by NiceGUI are not wanted in a particular styling.
        '''
        def str_to_dict(s: Optional[str]) -> Dict[str, str]:
            return {prop.split('=')[0]: prop.split('=')[1] if '=' in prop else True for prop in s.split()} if s else {}
        needs_update = False
        for key in str_to_dict(remove):
            if getattr(self.view, key, None) is not None:
                needs_update = True
            setattr(self.view, key, None)
        for key, value in str_to_dict(add).items():
            if getattr(self.view, key, None) != value:
                needs_update = True
            setattr(self.view, key, value)
        if needs_update:
            self.update()
        return self

    def tooltip(self, text: str, *, props: str = ''):
        tooltip = jp.QTooltip(text=text, temp=False)
        for prop in props.split():
            if '=' in prop:
                setattr(tooltip, *prop.split('='))
            else:
                setattr(tooltip, prop, True)
        tooltip.add_page(self.page)
        self.view.add(tooltip)
        self.update()
        return self

    def update(self) -> None:
        if globals.loop is not None:
            create_task(self.view.update())



import traceback
from dataclasses import dataclass
from inspect import signature
from typing import Any, Callable, List, Optional

from starlette.websockets import WebSocket

from . import globals
from .elements.element import Element
from .helpers import is_coroutine
from .lifecycle import on_startup
from .task_logger import create_task


@dataclass
class EventArguments:
    sender: Element
    socket: Optional[WebSocket]


@dataclass
class ClickEventArguments(EventArguments):
    pass


@dataclass
class ColorPickEventArguments(EventArguments):
    color: str


@dataclass
class MouseEventArguments(EventArguments):
    type: str
    image_x: float
    image_y: float


@dataclass
class UploadEventArguments(EventArguments):
    files: List[bytes]


@dataclass
class ValueChangeEventArguments(EventArguments):
    value: Any


@dataclass
class KeyboardAction:
    keydown: bool
    keyup: bool
    repeat: bool


@dataclass
class KeyboardModifiers:
    alt: bool
    ctrl: bool
    meta: bool
    shift: bool


@dataclass
class KeyboardKey:
    name: str
    code: str
    location: int

    def __eq__(self, other: str) -> bool:
        return self.name == other or self.code == other

    def __repr__(self):
        return str(self.name)

    @property
    def is_cursorkey(self):
        return self.code.startswith('Arrow')

    @property
    def number(self) -> Optional[int]:
        """Integer value of a number key."""
        return int(self.code.removeprefix('Digit')) if self.code.startswith('Digit') else None

    @property
    def backspace(self) -> bool:
        return self.name == 'Backspace'

    @property
    def tab(self) -> bool:
        return self.name == 'Tab'

    @property
    def enter(self) -> bool:
        return self.name == 'enter'

    @property
    def shift(self) -> bool:
        return self.name == 'Shift'

    @property
    def control(self) -> bool:
        return self.name == 'Control'

    @property
    def alt(self) -> bool:
        return self.name == 'Alt'

    @property
    def pause(self) -> bool:
        return self.name == 'Pause'

    @property
    def caps_lock(self) -> bool:
        return self.name == 'CapsLock'

    @property
    def escape(self) -> bool:
        return self.name == 'Escape'

    @property
    def space(self) -> bool:
        return self.name == 'Space'

    @property
    def page_up(self) -> bool:
        return self.name == 'PageUp'

    @property
    def page_down(self) -> bool:
        return self.name == 'PageDown'

    @property
    def end(self) -> bool:
        return self.name == 'End'

    @property
    def home(self) -> bool:
        return self.name == 'Home'

    @property
    def arrow_left(self) -> bool:
        return self.name == 'ArrowLeft'

    @property
    def arrow_up(self) -> bool:
        return self.name == 'ArrowUp'

    @property
    def arrow_right(self) -> bool:
        return self.name == 'ArrowRight'

    @property
    def arrow_down(self) -> bool:
        return self.name == 'ArrowDown'

    @property
    def print_screen(self) -> bool:
        return self.name == 'PrintScreen'

    @property
    def insert(self) -> bool:
        return self.name == 'Insert'

    @property
    def delete(self) -> bool:
        return self.name == 'Delete'

    @property
    def meta(self) -> bool:
        return self.name == 'Meta'

    @property
    def f1(self) -> bool:
        return self.name == 'F1'

    @property
    def f2(self) -> bool:
        return self.name == 'F2'

    @property
    def f3(self) -> bool:
        return self.name == 'F3'

    @property
    def f4(self) -> bool:
        return self.name == 'F4'

    @property
    def f5(self) -> bool:
        return self.name == 'F5'

    @property
    def f6(self) -> bool:
        return self.name == 'F6'

    @property
    def f7(self) -> bool:
        return self.name == 'F7'

    @property
    def f8(self) -> bool:
        return self.name == 'F8'

    @property
    def f9(self) -> bool:
        return self.name == 'F9'

    @property
    def f10(self) -> bool:
        return self.name == 'F10'

    @property
    def f11(self) -> bool:
        return self.name == 'F11'

    @property
    def f12(self) -> bool:
        return self.name == 'F12'


@dataclass
class KeyEventArguments(EventArguments):
    action: KeyboardAction
    key: KeyboardKey
    modifiers: KeyboardModifiers


def handle_event(handler: Optional[Callable], arguments: EventArguments) -> Optional[bool]:
    try:
        if handler is None:
            return False
        no_arguments = not signature(handler).parameters
        with globals.within_view(arguments.sender.parent_view):
            result = handler() if no_arguments else handler(arguments)
        if is_coroutine(handler):
            async def wait_for_result():
                with globals.within_view(arguments.sender.parent_view):
                    await result
            if globals.loop and globals.loop.is_running():
                create_task(wait_for_result(), name=str(handler))
            else:
                on_startup(None, wait_for_result())
        return False
    except Exception:
        traceback.print_exc()



import justpy as jp


class CustomView(jp.JustpyBaseComponent):

    def __init__(self, vue_type, **options):
        self.vue_type = vue_type

        self.pages = {}
        self.classes = ''
        self.style = ''
        self.options = jp.Dict(**options)
        self.components = []

        super().__init__(temp=False)

    def react(self, _):
        pass

    def convert_object_to_dict(self):
        return {
            'vue_type': self.vue_type,
            'id': self.id,
            'show': True,
            'classes': self.classes,
            'style': self.style,
            'options': self.options,
        }



import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, Awaitable, Callable, Optional

import justpy as jp
from starlette.requests import Request

from .task_logger import create_task

if TYPE_CHECKING:
    from .page import Page


@dataclass
class PageBuilder:
    function: Callable[[], Awaitable['Page']]
    shared: bool

    _shared_page: Optional['Page'] = None

    async def build(self) -> None:
        assert self.shared
        self._shared_page = await self.function()

    async def route_function(self, request: Request) -> 'Page':
        if self.shared:
            while self._shared_page is None:
                await asyncio.sleep(0.05)
            return self._shared_page
        else:
            page = await self.function(request)
        return await page._route_function(request)

    def create_route(self, route: str) -> None:
        if self.shared:
            create_task(self.build())
        jp.Route(route, self.route_function)



import base64
import traceback
from typing import Callable, Optional

import justpy as jp

from ..events import UploadEventArguments, handle_event
from .element import Element


class Upload(Element):

    def __init__(self, *,
                 multiple: bool = False,
                 on_upload: Optional[Callable] = None,
                 upload_button_text: str = 'Upload') -> None:
        """File Upload

        :param multiple: allow uploading multiple files at once (default: `False`)
        :param on_upload: callback to execute when a file is uploaded (list of bytearrays)
        :param upload_button_text: text for the upload button
        """
        self.upload_handler = on_upload
        view = jp.Form(
            enctype='multipart/form-data',
            classes='flex gap-4 items-center',
            submit=lambda sender, msg: self.submit(sender, msg),
            temp=False,
        )
        jp.Input(type='file', multiple=multiple, change=lambda *_: False, a=view, temp=False)
        jp.QBtn(type='submit', text=upload_button_text, color='primary', a=view, temp=False)

        super().__init__(view)

    def submit(self, _, msg) -> Optional[bool]:
        try:
            page_update = False
            for form_data in msg.form_data:
                if form_data.type == 'file':
                    files = [base64.b64decode(f.file_content) for f in form_data.files]
                    arguments = UploadEventArguments(sender=self, socket=msg.websocket, files=files)
                    if handle_event(self.upload_handler, arguments):
                        page_update = None
            return page_update
        except Exception:
            traceback.print_exc()



# isort:skip_file
import os


class Ui:
    from .run import run  # NOTE: before justpy

    from .page import page, add_head_html, add_body_html, run_javascript, await_javascript
    from .update import update

    from .elements.button import Button as button
    from .elements.card import Card as card
    from .elements.card import CardSection as card_section
    from .elements.chart import Chart as chart
    from .elements.checkbox import Checkbox as checkbox
    from .elements.color_input import ColorInput as color_input
    from .elements.color_picker import ColorPicker as color_picker
    from .elements.colors import Colors as colors
    from .elements.column import Column as column
    from .elements.custom_example import CustomExample as custom_example
    from .elements.dialog import Dialog as dialog
    from .elements.expansion import Expansion as expansion
    from .elements.html import Html as html
    from .elements.icon import Icon as icon
    from .elements.image import Image as image
    from .elements.input import Input as input
    from .elements.interactive_image import InteractiveImage as interactive_image
    from .elements.joystick import Joystick as joystick
    from .elements.keyboard import Keyboard as keyboard
    from .elements.label import Label as label
    from .elements.link import Link as link
    from .elements.log import Log as log
    from .elements.markdown import Markdown as markdown
    from .elements.menu import Menu as menu
    from .elements.menu_item import MenuItem as menu_item
    from .elements.menu_separator import MenuSeparator as menu_separator
    from .elements.notify import Notify as notify
    from .elements.number import Number as number
    from .elements.open import open, open_async
    from .elements.radio import Radio as radio
    from .elements.row import Row as row
    from .elements.scene import Scene as scene
    from .elements.select import Select as select
    from .elements.slider import Slider as slider
    from .elements.switch import Switch as switch
    from .elements.table import Table as table
    from .elements.toggle import Toggle as toggle
    from .elements.tree import Tree as tree
    from .elements.upload import Upload as upload
    from .lifecycle import on_connect, on_disconnect, on_shutdown, on_startup, shutdown
    from .routes import add_route, add_static_files, get
    from .timer import Timer as timer

    if os.environ.get('MATPLOTLIB', 'true').lower() == 'true':
        from .elements.line_plot import LinePlot as line_plot
        from .elements.plot import Plot as plot



from typing import Any

import justpy as jp

from .element import Element


class Html(Element):

    def __init__(self, content: str = ''):
        """HTML Element

        Renders arbitrary HTML onto the page. `Tailwind <https://tailwindcss.com/>`_ can be used for styling.
        You can also use `ui.add_head_html` to add html code into the head of the document and `ui.add_body_html`
        to add it into the body.

        :param content: the HTML code to be displayed
        """
        view = jp.QDiv(temp=False)
        super().__init__(view)
        self.content = content

    @property
    def content(self) -> str:
        return self.view.inner_html

    @content.setter
    def content(self, content: str) -> None:
        self.set_content(content)

    def set_content(self, content: str) -> None:
        if '</script>' in content:
            raise ValueError('HTML elements must not contain <script> tags. Use ui.add_body_html() instead.')
        self.view.inner_html = content
        self.update()



from typing import Dict

import justpy as jp

from .element import Element


class Chart(Element):

    def __init__(self, options: Dict):
        """Chart

        An element to create a chart using `Highcharts <https://www.highcharts.com/>`_.

        :param options: dictionary of Highcharts options
        """
        view = jp.HighCharts(temp=False)
        view.options = self.options = jp.Dict(**options)
        super().__init__(view)



from typing import Callable, Union

import justpy as jp

from .. import globals
from .group import Group


class Link(Group):

    def __init__(self, text: str = '', target: Union[Callable, str] = '#'):
        """Link

        Create a hyperlink.

        :param text: display text
        :param target: page function or string that is a an absolute URL or relative path from base URL
        """
        href = target if isinstance(target, str) else globals.find_route(target)[1:]
        view = jp.A(text=text, href=href, classes='underline text-blue', temp=False)

        super().__init__(view)



from typing import Callable, Optional

import justpy as jp

from .value_element import ValueElement


class FloatElement(ValueElement):

    def __init__(self, view: jp.HTMLBaseComponent, *, value: float, format: str = None, on_change: Optional[Callable]):
        self.format = format

        super().__init__(view, value=value, on_change=on_change)

    def value_to_view(self, value: float):
        if value is None:
            return None
        elif self.format is None:
            return str(value)
        else:
            return self.format % float(value)



from typing import Any, Callable, Dict, List, Optional, Union

import justpy as jp

from .value_element import ValueElement


class ChoiceElement(ValueElement):

    def __init__(self, view: jp.HTMLBaseComponent, options: Union[List, Dict], *,
                 value: Any, on_change: Optional[Callable] = None):
        if isinstance(options, List):
            view.options = [{'label': option, 'value': option} for option in options]
        else:
            view.options = [{'label': value, 'value': key} for key, value in options.items()]

        super().__init__(view, value=value, on_change=on_change)



from typing import Any, Callable, Dict, Optional

import justpy as jp

from ..binding import BindableProperty, bind_from, bind_to
from ..events import ValueChangeEventArguments, handle_event
from .element import Element


class ValueElement(Element):
    value = BindableProperty(on_change=lambda sender, value: handle_event(
        sender.change_handler, ValueChangeEventArguments(sender=sender, socket=None, value=value)))

    def __init__(self, view: jp.HTMLBaseComponent, *, value: Any, on_change: Optional[Callable]):
        super().__init__(view)

        self.change_handler = on_change
        self.value = value
        self.bind_value_to(self.view, 'value', forward=self.value_to_view)

    def value_to_view(self, value):
        return value

    def handle_change(self, msg: Dict):
        self.value = msg['value']
        self.update()
        return False

    def bind_value_to(self, target_object, target_name, *, forward=lambda x: x):
        bind_to(self, 'value', target_object, target_name, forward=forward)
        return self

    def bind_value_from(self, target_object, target_name, *, backward=lambda x: x):
        bind_from(self, 'value', target_object, target_name, backward=backward)
        return self

    def bind_value(self, target_object, target_name, *, forward=lambda x: x, backward=lambda x: x):
        bind_from(self, 'value', target_object, target_name, backward=backward)
        bind_to(self, 'value', target_object, target_name, forward=forward)
        return self



from typing import Any, Callable, Dict, List, Optional, Union

import justpy as jp

from .choice_element import ChoiceElement


class Radio(ChoiceElement):

    def __init__(self, options: Union[List, Dict], *, value: Any = None, on_change: Optional[Callable] = None):
        """Radio Selection

        :param options: a list ['value1', ...] or dictionary `{'value1':'label1', ...}` specifying the options
        :param value: the initial value
        :param on_change: callback to execute when selection changes
        """
        view = jp.QOptionGroup(options=options, input=self.handle_change, temp=False)

        super().__init__(view, options, value=value, on_change=on_change)



from typing import Callable, Optional

import justpy as jp

from .value_element import ValueElement


class BoolElement(ValueElement):

    def __init__(self, view: jp.HTMLBaseComponent, *, value: bool, on_change: Optional[Callable]):
        super().__init__(view, value=value, on_change=on_change)



from __future__ import annotations

import asyncio
import inspect
import time
import uuid
from functools import wraps
from typing import Callable, Optional

import justpy as jp
import starlette
from addict import Dict
from pygments.formatters import HtmlFormatter

from . import globals
from .helpers import is_coroutine
from .page_builder import PageBuilder


class Page(jp.QuasarPage):

    def __init__(self,
                 title: Optional[str] = None,
                 *,
                 favicon: Optional[str] = None,
                 dark: Optional[bool] = ...,
                 classes: str = 'q-ma-md column items-start',
                 css: str = HtmlFormatter().get_style_defs('.codehilite'),
                 on_connect: Optional[Callable] = None,
                 on_page_ready: Optional[Callable] = None,
                 on_disconnect: Optional[Callable] = None,
                 shared: bool = False,
                 ):
        super().__init__()

        if globals.config:
            self.title = title or globals.config.title
            self.favicon = favicon or globals.config.favicon
            self.dark = dark if dark is not ... else globals.config.dark
        else:
            self.title = title
            self.favicon = favicon
            self.dark = dark if dark is not ... else None
        self.tailwind = True  # use Tailwind classes instead of Quasars
        self.css = css
        self.connect_handler = on_connect
        self.page_ready_handler = on_page_ready
        self.disconnect_handler = on_disconnect
        self.delete_flag = not shared

        self.waiting_javascript_commands: dict[str, str] = {}
        self.on('result_ready', self.handle_javascript_result)
        self.on('page_ready', self.handle_page_ready)

        self.view = jp.Div(a=self, classes=classes, style='row-gap: 1em', temp=False)
        self.view.add_page(self)

    async def _route_function(self, request: starlette.requests.Request) -> Page:
        for handler in globals.connect_handlers + ([self.connect_handler] if self.connect_handler else []):
            arg_count = len(inspect.signature(handler).parameters)
            is_coro = is_coroutine(handler)
            if arg_count == 1:
                await handler(request) if is_coro else handler(request)
            elif arg_count == 0:
                await handler() if is_coro else handler()
            else:
                raise ValueError(f'invalid number of arguments (0 or 1 allowed, got {arg_count})')
        return self

    async def handle_page_ready(self, msg: Dict) -> bool:
        if self.page_ready_handler:
            arg_count = len(inspect.signature(self.page_ready_handler).parameters)
            is_coro = is_coroutine(self.page_ready_handler)
            if arg_count == 1:
                await self.page_ready_handler(msg.websocket) if is_coro else self.page_ready_handler(msg.websocket)
            elif arg_count == 0:
                await self.page_ready_handler() if is_coro else self.page_ready_handler()
            else:
                raise ValueError(f'invalid number of arguments (0 or 1 allowed, got {arg_count})')
        return False

    async def on_disconnect(self, websocket=None) -> None:
        for handler in globals.disconnect_handlers + ([self.disconnect_handler] if self.disconnect_handler else[]):
            arg_count = len(inspect.signature(handler).parameters)
            is_coro = is_coroutine(handler)
            if arg_count == 1:
                await handler(websocket) if is_coro else handler(websocket)
            elif arg_count == 0:
                await handler() if is_coro else handler()
            else:
                raise ValueError(f'invalid number of arguments (0 or 1 allowed, got {arg_count})')
        await super().on_disconnect(websocket)

    async def await_javascript(self, code: str, *, check_interval: float = 0.01, timeout: float = 1.0) -> str:
        start_time = time.time()
        request_id = str(uuid.uuid4())
        await self.run_javascript(code, request_id=request_id)
        while request_id not in self.waiting_javascript_commands:
            if time.time() > start_time + timeout:
                raise TimeoutError('JavaScript did not respond in time')
            await asyncio.sleep(check_interval)
        return self.waiting_javascript_commands.pop(request_id)

    def handle_javascript_result(self, msg) -> bool:
        self.waiting_javascript_commands[msg.request_id] = msg.result
        return False


def add_head_html(self, html: str) -> None:
    for page in find_parent_view().pages.values():
        page.head_html += html


def add_body_html(self, html: str) -> None:
    for page in find_parent_view().pages.values():
        page.body_html += html


async def run_javascript(self, code: str) -> None:
    for page in find_parent_view().pages.values():
        await page.run_javascript(code)


async def await_javascript(self, code: str, *, check_interval: float = 0.01, timeout: float = 1.0) -> None:
    for page in find_parent_view().pages.values():
        return await page.await_javascript(code)


class page:
    def __init__(
        self,
        route: str,
        title: Optional[str] = None,
        *,
        favicon: Optional[str] = None,
        dark: Optional[bool] = ...,
        classes: str = 'q-ma-md column items-start',
        css: str = HtmlFormatter().get_style_defs('.codehilite'),
        on_connect: Optional[Callable] = None,
        on_page_ready: Optional[Callable] = None,
        on_disconnect: Optional[Callable] = None,
        shared: bool = False,
    ):
        """Page

        Creates a new page at the given route.

        :param route: route of the new page (path must start with '/')
        :param title: optional page title
        :param favicon: optional favicon
        :param dark: whether to use Quasar's dark mode (defaults to `dark` argument of `run` command)
        :param classes: tailwind classes for the container div (default: `'q-ma-md column items-start'`)
        :param css: CSS definitions
        :param on_connect: optional function or coroutine which is called for each new client connection
        :param on_page_ready: optional function or coroutine which is called when the websocket is connected
        :param on_disconnect: optional function or coroutine which is called when a client disconnects
        :param shared: whether the page instance is shared between multiple clients (default: `False`)
        """
        self.route = route
        self.title = title
        self.favicon = favicon
        self.dark = dark
        self.classes = classes
        self.css = css
        self.on_connect = on_connect
        self.on_page_ready = on_page_ready
        self.on_disconnect = on_disconnect
        self.shared = shared
        self.page: Optional[Page] = None

    def __call__(self, func, *args, **kwargs):

        @wraps(func)
        async def decorated(request: starlette.requests.Request = None):
            self.page = Page(
                title=self.title,
                favicon=self.favicon,
                dark=self.dark,
                classes=self.classes,
                css=self.css,
                on_connect=self.on_connect,
                on_page_ready=self.on_page_ready,
                on_disconnect=self.on_disconnect,
                shared=self.shared,
            )
            with globals.within_view(self.page.view):
                if 'request' in inspect.signature(func).parameters:
                    if self.shared:
                        globals.log.error('cannot use `request` argument in shared page; providing 404 error page')
                        return error404()
                    kwargs['request'] = request
                await self.connected(request)
                await self.header()
                await func(*args, **kwargs) if is_coroutine(func) else func(*args, **kwargs)
                await self.footer()
            return self.page
        builder = PageBuilder(decorated, self.shared)
        if globals.server:
            builder.create_route(self.route)
        globals.page_builders[self.route] = builder
        return decorated

    async def connected(self, request: starlette.requests.Request):
        pass

    async def header(self):
        pass

    async def footer(self):
        pass


def find_parent_view() -> jp.HTMLBaseComponent:
    view_stack = globals.get_view_stack()
    if not view_stack:
        if globals.loop and globals.loop.is_running():
            raise RuntimeError('cannot find parent view, view stack is empty')
        page = Page(shared=True)
        view_stack.append(page.view)
        jp.Route('/', page._route_function)
    return view_stack[-1]


def error404() -> jp.QuasarPage:
    title = globals.config.title if globals.config else '404'
    favicon = globals.config.favicon if globals.config else None
    dark = globals.config.dark if globals.config else False
    wp = jp.QuasarPage(title=title, favicon=favicon, dark=dark, tailwind=True)
    div = jp.Div(a=wp, classes='py-20 text-center')
    jp.Div(a=div, classes='text-8xl py-5', text='☹',
           style='font-family: "Arial Unicode MS", "Times New Roman", Times, serif;')
    jp.Div(a=div, classes='text-6xl py-5', text='404')
    jp.Div(a=div, classes='text-xl py-5', text='This page doesn\'t exist.')
    return wp


def init_auto_index_page() -> None:
    view_stack = globals.view_stacks.get(0)
    if not view_stack:
        return  # there is no auto-index page on the view stack
    page: Page = view_stack.pop().pages[0]
    page.title = globals.config.title
    page.favicon = globals.config.favicon
    page.dark = globals.config.dark
    page.view.classes = globals.config.main_page_classes
    assert len(view_stack) == 0


def create_page_routes() -> None:
    jp.Route("/{path:path}", error404, last=True)
    for route, page_builder in globals.page_builders.items():
        page_builder.create_route(route)



# isort:skip_file
import asyncio
from typing import Awaitable, Callable

if True:  # NOTE: prevent formatter from mixing up these lines
    import builtins
    print_backup = builtins.print
    builtins.print = lambda *args, **kwargs: kwargs.get('flush') and print_backup(*args, **kwargs)
    from .ui import Ui  # NOTE: before justpy
    import justpy as jp
    builtins.print = print_backup

from . import binding, globals
from .page import create_page_routes, init_auto_index_page
from .task_logger import create_task
from .routes import create_exclude_routes
from .timer import Timer

jp.app.router.on_startup.clear()  # NOTE: remove JustPy's original startup function


@jp.app.on_event('startup')
async def patched_justpy_startup():
    jp.WebPage.loop = jp.asyncio.get_event_loop()
    jp.JustPy.loop = jp.WebPage.loop
    jp.JustPy.STATIC_DIRECTORY = jp.os.environ["STATIC_DIRECTORY"]
    print(f'NiceGUI ready to go on {"https" if jp.SSL_KEYFILE else "http"}://{jp.HOST}:{jp.PORT}')


@jp.app.on_event('startup')
def startup():
    globals.loop = asyncio.get_running_loop()
    init_auto_index_page()
    create_page_routes()
    create_exclude_routes()
    globals.tasks.extend(create_task(t.coro, name=t.name) for t in Timer.prepared_coroutines)
    Timer.prepared_coroutines.clear()
    globals.tasks.extend(create_task(t, name='startup task')
                         for t in globals.startup_handlers if isinstance(t, Awaitable))
    [safe_invoke(t) for t in globals.startup_handlers if isinstance(t, Callable)]
    jp.run_task(binding.loop())


@jp.app.on_event('shutdown')
def shutdown():
    [create_task(t, name='shutdown task') for t in globals.shutdown_handlers if isinstance(t, Awaitable)]
    [safe_invoke(t) for t in globals.shutdown_handlers if isinstance(t, Callable)]
    [t.cancel() for t in globals.tasks]


def safe_invoke(func: Callable):
    try:
        result = func()
        if isinstance(result, Awaitable):
            create_task(result)
    except:
        globals.log.exception(f'could not invoke {func}')


app = globals.app = jp.app
ui = Ui()



import inspect
import logging
import os
import sys
import webbrowser
from typing import List, Optional

import uvicorn
from uvicorn.main import STARTUP_FAILURE
from uvicorn.supervisors import ChangeReload, Multiprocess

from . import globals
from .config import Config

os.environ['STATIC_DIRECTORY'] = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'static')
os.environ['TEMPLATES_DIRECTORY'] = os.path.join(os.environ['STATIC_DIRECTORY'], 'templates')


def run(self, *,
        host: str = os.environ.get('HOST', '0.0.0.0'),
        port: int = int(os.environ.get('PORT', '8080')),
        title: str = 'NiceGUI',
        favicon: str = 'favicon.ico',
        dark: Optional[bool] = False,
        main_page_classes: str = 'q-ma-md column items-start',
        binding_refresh_interval: float = 0.1,
        show: bool = True,
        reload: bool = True,
        uvicorn_logging_level: str = 'warning',
        uvicorn_reload_dirs: str = '.',
        uvicorn_reload_includes: str = '*.py',
        uvicorn_reload_excludes: str = '.*, .py[cod], .sw.*, ~*',
        exclude: str = '',
        ):
    globals.config = Config(
        host=host,
        port=port,
        title=title,
        reload=reload,
        favicon=favicon,
        dark=dark,
        main_page_classes=main_page_classes,
        binding_refresh_interval=binding_refresh_interval,
        excludes=[e.strip() for e in exclude.split(',')],
    )
    os.environ['HOST'] = globals.config.host
    os.environ['PORT'] = str(globals.config.port)

    if inspect.stack()[-2].filename.endswith('spawn.py'):
        return

    if show:
        webbrowser.open(f'http://{host if host != "0.0.0.0" else "127.0.0.1"}:{port}/')

    def split_args(args: str) -> List[str]:
        return args.split(',') if ',' in args else [args]

    # NOTE: The following lines are basically a copy of `uvicorn.run`, but keep a reference to the `server`.

    config = uvicorn.Config(
        'nicegui:app' if reload else globals.app,
        host=host,
        port=port,
        lifespan='on',
        reload=reload,
        reload_includes=split_args(uvicorn_reload_includes) if reload else None,
        reload_excludes=split_args(uvicorn_reload_excludes) if reload else None,
        reload_dirs=split_args(uvicorn_reload_dirs) if reload else None,
        log_level=uvicorn_logging_level,
    )
    globals.server = uvicorn.Server(config=config)

    if (reload or config.workers > 1) and not isinstance(config.app, str):
        logging.warning('You must pass the application as an import string to enable "reload" or "workers".')
        sys.exit(1)

    if config.should_reload:
        sock = config.bind_socket()
        ChangeReload(config, target=globals.server.run, sockets=[sock]).run()
    elif config.workers > 1:
        sock = config.bind_socket()
        Multiprocess(config, target=globals.server.run, sockets=[sock]).run()
    else:
        globals.server.run()
    if config.uds:
        os.remove(config.uds)  # pragma: py-win32

    if not globals.server.started and not config.should_reload and config.workers == 1:
        sys.exit(STARTUP_FAILURE)



import traceback
from dataclasses import dataclass
from typing import Callable, Optional

import websockets
from justpy import WebPage

from .. import globals
from ..events import handle_event
from ..page import Page
from ..routes import add_dependencies
from ..task_logger import create_task
from .custom_view import CustomView
from .element import Element
from .scene_object3d import Object3D

add_dependencies(__file__, [
    'three.min.js',
    'CSS2DRenderer.js',
    'CSS3DRenderer.js',
    'OrbitControls.js',
    'STLLoader.js',
    'tween.umd.min.js',
])


@dataclass
class SceneCamera:
    x: float = 0
    y: float = -3
    z: float = 5
    look_at_x: float = 0
    look_at_y: float = 0
    look_at_z: float = 0
    up_x: float = 0
    up_y: float = 0
    up_z: float = 1

    def create_move_command(self, duration: float = 0) -> str:
        return 'move_camera(' \
            f'{self.x}, {self.y}, {self.z}, ' \
            f'{self.look_at_x}, {self.look_at_y}, {self.look_at_z}, ' \
            f'{self.up_x}, {self.up_y}, {self.up_z}, {duration})'


class SceneView(CustomView):

    def __init__(self, *, width: int, height: int, on_click: Optional[Callable]):
        super().__init__('scene', width=width, height=height)
        self.on_click = on_click
        self.allowed_events = ['onConnect', 'onClick']
        self.initialize(temp=False, onConnect=self.handle_connect, onClick=self.handle_click)
        self.objects = {}
        self.camera: SceneCamera = SceneCamera()

    def handle_connect(self, msg):
        try:
            for object in self.objects.values():
                object.send_to(msg.websocket)
            create_task(self.run_method(self.camera.create_move_command(), msg.websocket), name='move camera (connect)')
        except:
            traceback.print_exc()

    def handle_click(self, msg) -> Optional[bool]:
        try:
            for hit in msg.hits:
                hit.object = self.objects.get(hit.object_id)
            return handle_event(self.on_click, msg)
        except:
            traceback.print_exc()

    async def run_method(self, command, websocket):
        try:
            await websocket.send_json({'type': 'run_method', 'data': command, 'id': self.id})
        except (websockets.exceptions.ConnectionClosedOK, RuntimeError):
            pass
        return True

    def __len__(self):
        return len(self.objects)


class Scene(Element):
    from .scene_objects import Box as box
    from .scene_objects import Curve as curve
    from .scene_objects import Cylinder as cylinder
    from .scene_objects import Extrusion as extrusion
    from .scene_objects import Group as group
    from .scene_objects import Line as line
    from .scene_objects import QuadraticBezierTube as quadratic_bezier_tube
    from .scene_objects import Ring as ring
    from .scene_objects import Sphere as sphere
    from .scene_objects import SpotLight as spot_light
    from .scene_objects import Stl as stl
    from .scene_objects import Text as text
    from .scene_objects import Text3d as text3d
    from .scene_objects import Texture as texture

    def __init__(self, width: int = 400, height: int = 300, on_click: Optional[Callable] = None):
        """3D Scene

        Display a 3d scene using `three.js <https://threejs.org/>`_.
        Currently NiceGUI supports boxes, spheres, cylinders/cones, extrusions, straight lines, curves and textured meshes.
        Objects can be translated, rotated and displayed with different color, opacity or as wireframes.
        They can also be grouped to apply joint movements.

        :param width: width of the canvas
        :param height: height of the canvas
        :param on_click: callback to execute when a 3d object is clicked
        """
        super().__init__(SceneView(width=width, height=height, on_click=on_click))

    def __enter__(self):
        globals.get_view_stack().append(self.view)
        scene = self.view.objects.get('scene', SceneObject(self.view, self.page))
        Object3D.stack.clear()
        Object3D.stack.append(scene)
        return self

    def __exit__(self, *_):
        globals.get_view_stack().pop()

    def move_camera(self,
                    x: Optional[float] = None,
                    y: Optional[float] = None,
                    z: Optional[float] = None,
                    look_at_x: Optional[float] = None,
                    look_at_y: Optional[float] = None,
                    look_at_z: Optional[float] = None,
                    up_x: Optional[float] = None,
                    up_y: Optional[float] = None,
                    up_z: Optional[float] = None,
                    duration: float = 0.5):
        camera: SceneCamera = self.view.camera
        camera.x = camera.x if x is None else x
        camera.y = camera.y if y is None else y
        camera.z = camera.z if z is None else z
        camera.look_at_x = camera.look_at_x if look_at_x is None else look_at_x
        camera.look_at_y = camera.look_at_y if look_at_y is None else look_at_y
        camera.look_at_z = camera.look_at_z if look_at_z is None else look_at_z
        camera.up_x = camera.up_x if up_x is None else up_x
        camera.up_y = camera.up_y if up_y is None else up_y
        camera.up_z = camera.up_z if up_z is None else up_z
        for socket in WebPage.sockets.get(self.page.page_id, {}).values():
            create_task(self.view.run_method(camera.create_move_command(duration), socket), name='move camera')


class SceneObject:

    def __init__(self, view: SceneView, page: Page):
        self.id = 'scene'
        self.view = view
        self.page = page



from typing import Callable, Optional

import justpy as jp

from .string_element import StringElement


class Input(StringElement):

    def __init__(self, label: str = None, *,
                 placeholder: str = None, value: str = '', on_change: Optional[Callable] = None):
        """Text Input

        :param label: displayed label for the text input
        :param placeholder: text to show if no value is entered
        :param value: the current value of the text input
        :param on_change: callback to execute when the input is confirmed by leaving the focus
        """
        view = jp.QInput(
            label=label,
            placeholder=placeholder,
            value=value,
            change=self.handle_change,
            disable_input_event=True,
            temp=False,
        )

        super().__init__(view, value=value, on_change=on_change)



from typing import List

from . import globals
from .elements.element import Element
from .task_logger import create_task


def update(self, *elements: List[Element]) -> None:
    if not (globals.loop and globals.loop.is_running()):
        return
    for element in elements:
        create_task(element.view.update())



import asyncio
import time
import traceback
from collections import namedtuple
from typing import Callable, List

from . import globals
from .binding import BindableProperty
from .helpers import is_coroutine
from .page import find_parent_view
from .task_logger import create_task

NamedCoroutine = namedtuple('NamedCoroutine', ['name', 'coro'])


class Timer:
    prepared_coroutines: List[NamedCoroutine] = []

    active = BindableProperty()
    interval = BindableProperty()

    def __init__(self, interval: float, callback: Callable, *, active: bool = True, once: bool = False):
        """Timer

        One major drive behind the creation of NiceGUI was the necessity to have a simple approach to update the interface in regular intervals, for example to show a graph with incomming measurements.
        A timer will execute a callback repeatedly with a given interval.

        :param interval: the interval in which the timer is called (can be changed during runtime)
        :param callback: function or coroutine to execute when interval elapses
        :param active: whether the callback should be executed or not (can be changed during runtime)
        :param once: whether the callback is only executed once after a delay specified by `interval` (default: `False`)
        """

        self.active = active
        self.interval = interval
        self.parent_view = find_parent_view()

        async def do_callback():
            try:
                with globals.within_view(self.parent_view):
                    result = callback()
                    if is_coroutine(callback):
                        await result
            except Exception:
                traceback.print_exc()

        async def timeout():
            await asyncio.sleep(self.interval)
            await do_callback()

        async def loop():
            while True:
                try:
                    start = time.time()
                    if self.active:
                        await do_callback()
                    dt = time.time() - start
                    await asyncio.sleep(self.interval - dt)
                except asyncio.CancelledError:
                    return
                except:
                    traceback.print_exc()
                    await asyncio.sleep(self.interval)

        coroutine = timeout() if once else loop()
        if not (globals.loop and globals.loop.is_running()):
            self.prepared_coroutines.append(NamedCoroutine(str(callback), coroutine))
        else:
            globals.tasks.append(create_task(coroutine, name=str(callback)))



from typing import Callable, Optional

import justpy as jp
from nicegui.events import ValueChangeEventArguments, handle_event

from .element import Element


class Tree(Element):
    def __init__(self, nodes: list, *,
                 node_key: str = 'id',
                 label_key: str = 'label',
                 children_key: str = 'children',
                 on_select: Optional[Callable] = None):
        """Tree

        :param nodes: hierarchical list of node objects
        :param node_key: property name of each node object that holds its unique id (default: "id")
        :param label_key: property name of each node object that holds its label (default: "label")
        :param children_key: property name of each node object that holds its list of children (default: "children")
        :param on_click: callback which is invoked when the node selection changes
        """

        view = jp.QTree(nodes=nodes, node_key=node_key, label_key=label_key, children_key=children_key, temp=False)
        super().__init__(view)

        def process_event(view, event) -> Optional[bool]:
            arguments = ValueChangeEventArguments(sender=self, socket=event.get('websocket'), value=event.get('value'))
            return handle_event(on_select, arguments)

        view.on('update:selected', process_event)
        view.on('update:expanded', lambda *_: self.update())
        view.on('update:ticked', lambda *_: self.update())



from typing import Callable, Optional

import justpy as jp

from .bool_element import BoolElement


class Checkbox(BoolElement):

    def __init__(self, text: str = '', *, value: bool = False, on_change: Optional[Callable] = None):
        """Checkbox

        :param text: the label to display next to the checkbox
        :param value: whether it should be checked initially (default: `False`)
        :param on_change: callback to execute when value changes
        """
        view = jp.QCheckbox(text=text, input=self.handle_change, temp=False)

        super().__init__(view, value=value, on_change=on_change)



import justpy as jp

from .element import Element


class Icon(Element):

    def __init__(self, name: str):
        """Icon

        Displays an icon.

        `Here <https://material.io/icons/>`_ is a reference of possible names.

        :param name: the name of the icon
        """
        view = jp.QIcon(name=name, classes=f'q-pt-xs', size='20px', temp=False)

        super().__init__(view)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nicegui', 'nicegui.elements', 'nicegui.static.templates']

package_data = \
{'': ['*'],
 'nicegui': ['static/*'],
 'nicegui.elements': ['lib/*'],
 'nicegui.static.templates': ['css/*',
                              'highcharts/*',
                              'js/*',
                              'local/*',
                              'local/fontawesome/css/*',
                              'local/fontawesome/webfonts/*',
                              'local/ionicons/*',
                              'local/materialdesignicons/iconfont/*',
                              'local/robotofont/*',
                              'local/tailwind/*']}

install_requires = \
['Pygments>=2.9.0,<3.0.0',
 'asttokens>=2.0.5,<3.0.0',
 'docutils>=0.17.1,<0.18.0',
 'httpx>=0.23.0,<0.24.0',
 'justpy>=0.2.3,<0.3.0',
 'markdown2>=2.4.3,<3.0.0',
 'matplotlib>=3.4.1,<4.0.0',
 'typing-extensions>=3.10.0',
 'uvicorn>=0.18.0,<0.19.0',
 'watchfiles>=0.16.1,<0.17.0',
 'websockets>=10.3,<11.0']

setup_kwargs = {
    'name': 'nicegui',
    'version': '0.9.4',
    'description': 'Web User Interface with Buttons, Dialogs, Markdown, 3D Scences and Plots',
    'long_description': '<img src="https://raw.githubusercontent.com/zauberzeug/nicegui/main/sceenshots/ui-elements.png" width="300" align="right">\n\n# NiceGUI\n\nNiceGUI is an easy-to-use, Python-based UI framework, which renders to the web browser.\nYou can create buttons, dialogs, markdown, 3D scenes, plots and much more.\n\nIt is great for micro web apps, dashboards, robotics projects, smart home solutions and similar use cases.\nYou can also use it in development, for example when tweaking/configuring a machine learning algorithm or tuning motor controllers.\n\nNiceGUI is available as [PyPI package](https://pypi.org/project/nicegui/), [Docker image](https://hub.docker.com/r/zauberzeug/nicegui) and on [GitHub](https://github.com/zauberzeug/nicegui).\n\n[![PyPI version](https://badge.fury.io/py/nicegui.svg)](https://pypi.org/project/nicegui/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/nicegui)](https://pypi.org/project/nicegui/)\n[![Docker Pulls](https://img.shields.io/docker/pulls/zauberzeug/nicegui)](https://hub.docker.com/r/zauberzeug/nicegui)<br />\n[![GitHub commit activity](https://img.shields.io/github/commit-activity/m/zauberzeug/nicegui)](https://github.com/zauberzeug/nicegui/graphs/commit-activity)\n[![GitHub issues](https://img.shields.io/github/issues/zauberzeug/nicegui)](https://github.com/zauberzeug/nicegui/issues)\n[![GitHub forks](https://img.shields.io/github/forks/zauberzeug/nicegui)](https://github.com/zauberzeug/nicegui/network)\n[![GitHub stars](https://img.shields.io/github/stars/zauberzeug/nicegui)](https://github.com/zauberzeug/nicegui/stargazers)\n[![GitHub license](https://img.shields.io/github/license/zauberzeug/nicegui)](https://github.com/zauberzeug/nicegui/blob/main/LICENSE)\n\n## Features\n\n- browser-based graphical user interface\n- implicit reload on code change\n- standard GUI elements like label, button, checkbox, switch, slider, input, file upload, ...\n- simple grouping with rows, columns, cards and dialogs\n- general-purpose HTML and markdown elements\n- powerful high-level elements to\n  - plot graphs and charts,\n  - render 3D scenes,\n  - get steering events via virtual joysticks\n  - annotate and overlay images\n  - interact with tables\n  - navigate foldable tree structures\n- built-in timer to refresh data in intervals (even every 10 ms)\n- straight-forward data binding to write even less code\n- notifications, dialogs and menus to provide state of the art user interaction\n- shared and individual web pages\n- ability to add custom routes and data responses\n- capture keyboard input for global shortcuts etc\n- customize look by defining primary, secondary and accent colors\n- live-cycle events and session data\n\n## Installation\n\n```bash\npython3 -m pip install nicegui\n```\n\n## Usage\n\nWrite your nice GUI in a file `main.py`:\n\n```python\nfrom nicegui import ui\n\nui.label(\'Hello NiceGUI!\')\nui.button(\'BUTTON\', on_click=lambda: ui.notify(\'button was pressed\'))\n\nui.run()\n```\n\nLaunch it with:\n\n```bash\npython3 main.py\n```\n\nThe GUI is now available through http://localhost:8080/ in your browser.\nNote: The script will automatically reload the page when you modify the code.\n\nFull documentation can be found at [https://nicegui.io](https://nicegui.io).\n\n## Configuration\n\nYou can call `ui.run()` with optional arguments:\n\n- `host` (default: `\'0.0.0.0\'`)\n- `port` (default: `8080`)\n- `title` (default: `\'NiceGUI\'`)\n- `favicon` (default: `\'favicon.ico\'`)\n- `dark`: whether to use Quasar\'s dark mode (default: `False`, use `None` for "auto" mode)\n- `main_page_classes`: configure Quasar classes of main page (default: `\'q-ma-md column items-start\'`)\n- `binding_refresh_interval`: time between binding updates (default: `0.1` seconds, bigger is more cpu friendly)\n- `show`: automatically open the ui in a browser tab (default: `True`)\n- `reload`: automatically reload the ui on file changes (default: `True`)\n- `uvicorn_logging_level`: logging level for uvicorn server (default: `\'warning\'`)\n- `uvicorn_reload_dirs`: string with comma-separated list for directories to be monitored (default is current working directory only)\n- `uvicorn_reload_includes`: string with comma-separated list of glob-patterns which trigger reload on modification (default: `\'.py\'`)\n- `uvicorn_reload_excludes`: string with comma-separated list of glob-patterns which should be ignored for reload (default: `\'.*, .py[cod], .sw.*, ~*\'`)\n- `exclude`: comma-separated string to exclude elements (with corresponding JavaScript libraries) to save bandwidth\n  (possible entries: chart, colors, custom_example, interactive_image, keyboard, log, joystick, scene, table)\n\nThe environment variables `HOST` and `PORT` can also be used to configure NiceGUI.\n\nTo avoid the potentially costly import of Matplotlib, you set the environment variable `MATPLOTLIB=false`.\nThis will make `ui.plot` and `ui.line_plot` unavailable.\n\nNote:\nThe parameter `exclude` from earlier versions of NiceGUI has been removed.\nLibraries are now automatically served on demand.\nAs a small caveat, the page will be reloaded if a new dependency is added dynamically, e.g. when adding a `ui.chart` only after pressing a button.\n\n## Docker\n\nYou can use our [multi-arch Docker image](https://hub.docker.com/repository/docker/zauberzeug/nicegui) for pain-free installation:\n\n```bash\ndocker run --rm -p 8888:8080 -v $(pwd):/app/ -it zauberzeug/nicegui:latest\n```\n\nThis will start the server at http://localhost:8888 with the code from your current directory.\nThe file containing your `ui.run(port=8080, ...)` command must be named `main.py`.\nCode modification triggers an automatic reload.\n\n## Why?\n\nWe like [Streamlit](https://streamlit.io/) but find it does [too much magic when it comes to state handling](https://github.com/zauberzeug/nicegui/issues/1#issuecomment-847413651).\nIn search for an alternative nice library to write simple graphical user interfaces in Python we discovered [JustPy](https://justpy.io/).\nWhile it is too "low-level HTML" for our daily usage, it provides a great basis for NiceGUI.\n\n## Documentation and Examples\n\nThe API reference is hosted at [https://nicegui.io](https://nicegui.io).\nIt is [implemented with NiceGUI itself](https://github.com/zauberzeug/nicegui/blob/main/main.py).\nYou may also have a look at [examples.py](https://github.com/zauberzeug/nicegui/tree/main/examples.py) for more demonstrations of what you can do with NiceGUI.\n\n## Abstraction\n\nNiceGUI is based on [JustPy](https://justpy.io/) which is based on the ASGI framework [Starlette](https://www.starlette.io/) and the ASGI webserver [Uvicorn](https://www.uvicorn.org/).\n\n## Deployment\n\nTo deploy your NiceGUI app, you will need to execute your `main.py` (or whichever file contains your `ui.run(...)`) on your server infrastructure.\nYou can either install the [NiceGUI python package via pip](https://pypi.org/project/nicegui/) on the server or use our [pre-built Docker image](https://hub.docker.com/r/zauberzeug/nicegui) which contains all necessary dependencies.\nFor example you can use this `docker run` command to start the script `main.py` in the current directory on port 80:\n\n```bash\ndocker run -p 80:8080 -v $(pwd)/:/app/ -d --restart always zauberzeug/nicegui:latest\n```\n\nThe example assumes `main.py` uses the port 8080 in the `ui.run` command (which is the default).\nThe `--restart always` makes sure the container is restarted if the app crashes or the server reboots.\nOf course this can also be written in a docker compose file:\n\n```yaml\nnicegui:\n  image: zauberzeug/nicegui:latest\n  restart: always\n  ports:\n    - 80:8080\n  volumes:\n    - ./:/app/\n```\n\nWhile it is possible to provide SSL certificates directly through NiceGUI (using [JustPy config](https://justpy.io/reference/configuration/)), we also like reverse proxies like [Traefik](https://doc.traefik.io/traefik/) or [NGINX](https://www.nginx.com/).\nSee our [docker-compose.yml](https://github.com/zauberzeug/nicegui/blob/main/docker-compose.yml) as an example.\n',
    'author': 'Zauberzeug GmbH',
    'author_email': 'info@zauberzeug.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zauberzeug/nicegui',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import asyncio
from typing import Any, Optional

import justpy as jp

from .group import Group


class Dialog(Group):

    def __init__(self, *, value: bool = False):
        """Dialog

        Creates a modal dialog.

        :param value: whether the dialog is already opened (default: `False`)
        """
        view = jp.QDialog(value=value, input=self._on_input, temp=False)

        self._submitted: Optional[asyncio.Event] = None
        self._result: Any = None

        super().__init__(view)

    def open(self) -> None:
        self.view.value = True
        self.update()

    def close(self) -> None:
        self.view.value = False
        self.update()

    def __await__(self):
        self._submitted = asyncio.Event()
        self._submitted.clear()
        self._result = None
        self.open()
        yield from self.view.update().__await__()
        yield from self._submitted.wait().__await__()
        self.close()
        return self._result

    def submit(self, result: Any) -> bool:
        self._result = result
        self._submitted.set()
        return False

    def _on_input(self, *_) -> bool:
        self._result = None
        if self._submitted is not None:
            self._submitted.set()
        self.update()
        return False



from typing import Callable, Optional, Union

from justpy import WebPage
from starlette.websockets import WebSocket

from .. import globals
from ..task_logger import create_task


def open(self, target: Union[Callable, str], socket: Optional[WebSocket] = None):
    """Open

    Can be used to programmatically trigger redirects for a specific client.

    Note that *all* clients (i.e. browsers) connected to the page will open the target URL *unless* a socket is specified.
    User events like button clicks provide such a socket.

    :param target: page function or string that is a an absolute URL or relative path from base URL
    :param socket: optional WebSocket defining the target client
    """
    create_task(open_async(self, target, socket), name='open_async')


async def open_async(self, target: Union[Callable, str], socket: Optional[WebSocket] = None):
    path = target if isinstance(target, str) else globals.find_route(target)[1:]
    sockets = [socket] if socket else [s for socket_dict in WebPage.sockets.values() for s in socket_dict.values()]
    for socket in sockets:
        if not path:
            path = ' '  # NOTE empty string seems to be transformed to "null" when sending
        await socket.send_json({'type': 'page_update', 'page_options': {'redirect': path}})



from typing import Callable, Dict

import justpy as jp
from nicegui.events import ColorPickEventArguments, handle_event

from ..task_logger import create_task
from .element import Element


class ColorPicker(Element):

    def __init__(self, *, on_pick: Callable, value: bool = False):
        """Color Picker

        :param on_pick: callback to execute when a color is picked
        :param value: whether the menu is already opened (default: `False`)
        """
        view = jp.parse_html('''
            <q-popup-proxy transition-show="scale" transition-hide="scale" name="popup" no-parent-event>
                <q-color name="color_input"/>
            </q-popup-proxy>''')

        def handle_pick(sender, msg: Dict):
            return handle_event(on_pick, ColorPickEventArguments(sender=self, socket=msg.websocket, color=msg.value))
        view.name_dict['color_input'].on('change', handle_pick)
        view.name_dict['color_input'].disable_input_event = True
        view.name_dict['popup'].value = value
        view.name_dict['popup'].on('input', lambda *_: self.update() or False)
        view.name_dict['popup'].on('show', lambda *_: self.update() or False)
        view.name_dict['popup'].on('hide', lambda *_: self.update() or False)

        super().__init__(view)

    def open(self):
        self.view.name_dict['popup'].value = True
        self.update()

    def close(self):
        self.view.name_dict['popup'].value = False
        self.update()
